static void read_samples_part1(struct bgen_samples *samples, uint32_t nsamples,
                               uint32_t *samples_max_len);
static void read_samples_part2(struct bgen_samples const *samples, uint32_t nsamples,
                               char *const sample_array, uint32_t samples_stride);
